/*
 *  pH-compare-profiles.c
 *
 *  Copyright (C) 2000, 2001  Anil Somayaji
 *
 *  A utility for comparing pH profiles.
 * 
 *  pH is a kernel extension for online process intrusion detection
 *  monitoring and response.
 *
 *  Licensed under the GNU GPL (see COPYING file), and as such has NO
 *  WARRANTY!  In particular, pH modifies the filesystem using
 *  low-level access routines.  Have a good backup, you could lose
 *  data through disk corruption!
 *
 *  $Id: pH-compare-profiles.c,v 1.5 2001/09/10 00:03:38 soma Exp $
 *
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>

#define PH_NUM_SYSCALLS 256   /* size of array */
#define NUMSYS PH_NUM_SYSCALLS
#define PH_EMPTY_SYSCALL 255  /* note: this value is used as the */

#define PH_MAX_SEQLEN 9
#define MLOOK 8
#define PH_MAX_DISK_FILENAME 256
#define PH_LOCALITY_WIN 128
#define PH_FILE_MAGIC_LEN 20

typedef unsigned char u8;

typedef u8 pH_seqflags;

typedef struct pH_disk_profile_data {
        int sequences;  /* # sequences that have been inserted */
                        /*   NOT the number of lookahead pairs */
        unsigned long last_mod_count; /* # syscalls since last modification */
        unsigned long train_count;      /* # syscalls seen during training */
        u8 empty[PH_NUM_SYSCALLS];
        pH_seqflags entry[PH_NUM_SYSCALLS][PH_NUM_SYSCALLS];
} pH_disk_profile_data;

typedef struct pH_disk_profile {
        char magic[PH_FILE_MAGIC_LEN];  /* file magic: identifier, version */
        int normal;
        int frozen;
        time_t normal_time;
        int length;
        unsigned long count;
        int anomalies;
        pH_disk_profile_data train, test;
        char filename[PH_MAX_DISK_FILENAME];
} pH_disk_profile;

typedef struct look_stats {
        int look;
        int total;
        int count;
        unsigned int stats[PH_NUM_SYSCALLS][PH_NUM_SYSCALLS];
} look_stats;

const char *PH_FILE_MAGIC="pH profile 0.18\n";

int compare_ints_rev(const void *x, const void *y)
{
        unsigned int *a = (unsigned int *) x;
        unsigned int *b = (unsigned int *) y;

        if (*a > *b)
                return -1;
        else if (*a < *b)
                return 1;
        else
                return 0;
}

int write_image(look_stats *freq, char *image_prefix)
{
        char buf[BUFSIZ];
        FILE *file;
        int i, j, x;
        int maxval = 255;
        unsigned char red, green, blue, pixel[3];
        int freq_list[freq->count];
        int freq_count;

        sprintf(buf, "%s-%d-dist.pnm", image_prefix, freq->look);
        file = fopen(buf, "w");
        
        if (!file)
                return -1;
        
        fprintf(file, "P6\n");
        fprintf(file, "%d %d\n", NUMSYS, NUMSYS);
        fprintf(file, "%d\n", maxval);

        freq_count = 0;

        for (i = 0; i < NUMSYS; i++) {
                for (j = 0; j < NUMSYS; j++) {
                        if (freq->stats[i][j]) {
                                x = freq->stats[i][j];
                                if (freq_count >= freq->count) {
                                        fprintf(stderr,
                                                "Count %d! too small!\n",
                                                freq->count);
                                } else {
                                        freq_list[freq_count] = x;
                                        freq_count++;
                                }
                                if (x > (freq->total / 2)) {
                                        red = (x * maxval) / freq->total;
                                        green = maxval - red;
                                        blue = green;
                                } else {
                                        red = (x * maxval) / freq->total;
                                        green = red;
                                        blue = maxval - red;
                                }
                                printf("%d %d %d %d : %hhd %hhd %hhd\n",
                                       x, i, j, freq->look,
                                       red, green, blue);
                        } else {
                                red = maxval;
                                green = maxval;
                                blue = maxval;
                        }
                        pixel[0] = red;
                        pixel[1] = green;
                        pixel[2] = blue;
                        fwrite(pixel, 1, 3, file);
                }
        }

        fclose(file);

        qsort(freq_list, freq_count, sizeof(int), compare_ints_rev);

        sprintf(buf, "%s-%d-freq.jgr", image_prefix, freq->look);
        file = fopen(buf, "w");
        
        if (!file)
                return -2;
        
        fprintf(file, "newgraph\n\n"
                "xaxis min 0 label : Lookahead Pair\n\n"
                "yaxis min 0 label : Frequency\n"
                "no_draw_hash_marks\n");

        fprintf(file, "\ncurve 1\n"
                "  linetype solid\n"
                "  marktype none\n"
                "  label : w = %d\n"
                "pts\n", freq->look);

        for (i = 0; i < freq_count; i++) {
                fprintf(file, "%d  %d\n", i + 1, freq_list[i]);
        }

        fclose(file);

        sprintf(buf, "jgraph %s-%d-freq.jgr > %s-%d-freq.ps", 
                image_prefix, freq->look,
                image_prefix, freq->look);

        system(buf);
        
        sprintf(buf, "%s-%d-freq.dat", image_prefix, freq->look);
        file = fopen(buf, "w");
        if (!file)
                return -3;
        for (i = 0; i < freq_count; i++) {
                fprintf(file, "%d  %d\n", i + 1, freq_list[i]);
        }
        fclose(file);

        return 0;
}

void usage(char *program)
{
        fprintf(stderr, "Usage: %s <image-prefix> <kernel-profile-1> ...\n",
                program);
}

int main(int argc, char *argv[])
{
        pH_disk_profile profile;
        pH_disk_profile_data *data;
        look_stats freq[MLOOK];
        int retval;
        int fd;
        int i,j,k,file;
        int num_profiles, num_normal;
        u8 entry;
        char *image_prefix;

        if (argc < 3) {
                usage(argv[0]);
                return -1;
        }

        image_prefix = argv[1];

        num_profiles = 0;
        num_normal = 0;

        for (i=0; i<MLOOK; i++) {
                freq[i].look = i + 2;
                freq[i].total = 0;
                freq[i].count = 0;
                for (j=0; j<NUMSYS; j++) {
                        for (k = 0; k<NUMSYS; k++) {
                                freq[i].stats[j][k] = 0;
                        }
                }
        }        

        for (file=2; file < argc; file++) {
                fd = open(argv[file], O_RDONLY, 0);

                if (fd < 0) {
                        fprintf(stderr, "Can't open %s\n", argv[file]);
                        continue;
                }
        
                retval = read(fd, &profile, sizeof(profile));
                close(fd);

                if (retval != sizeof(profile)) {
                        fprintf(stderr, "Only got %d bytes from %s, "
                                "but wanted %d\n", retval, argv[file],
                                sizeof(profile));
                        continue;
                }

                if (strcmp(profile.magic, PH_FILE_MAGIC)) {
                        fprintf(stderr, "Magic is wrong for %s\n",
                                argv[file]);
                        continue;
                }

                num_profiles++;

                if (profile.normal) {
                        data = &(profile.test);
                        num_normal++;
                } else {
                        data = &(profile.train);
                }

                for (i=0; i<NUMSYS; i++) {
                        for (j=0; j<NUMSYS; j++) {
                                entry = data->entry[i][j];
                                for (k=0; k<MLOOK; k++) {
                                        if ((entry >> k) & ~(~0 << 1)) {
                                                freq[k].stats[i][j]++;
                                                if (freq[k].stats[i][j] == 1)
                                                        freq[k].count++;
                                        }
                                }
                        }
                }
        }
        
        for (i = 0; i < MLOOK; i++) {
                freq[i].total = num_profiles;
                write_image(freq + i, image_prefix);
        }

        fprintf(stderr, "%d profiles\n%d normal\n",
                num_profiles, num_normal);
        
        return 0;
}
