/*
 *  pH-print-profile.c
 *
 *  Copyright (C) 2001  Anil Somayaji
 *
 *  A utility for editing the contents of a pH profile.  pH is a
 *  kernel extension for online process intrusion detection monitoring
 *  and response.
 *
 *  Licensed under the GNU GPL (see COPYING file), and as such has NO
 *  WARRANTY!  In particular, pH modifies the filesystem using
 *  low-level access routines.  Have a good backup, you could lose
 *  data through disk corruption!
 *
 *  $Id: pH-edit-profile.c,v 1.2 2001/09/30 03:17:21 soma Exp $ */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <time.h>

#define PH_NUM_SYSCALLS 256
#define PH_MAX_SEQLEN 9
#define PH_MAX_FILENAME 255
#define PH_LOCALITY_WIN 128
#define PH_FILE_MAGIC_LEN 20
#define PH_MAX_DISK_FILENAME 256

const char *PH_FILE_MAGIC="pH profile 0.18\n";

typedef unsigned char u8;

typedef u8 pH_seqflags;

typedef struct pH_disk_profile_data {
        int sequences;
        unsigned long last_mod_count;/* # syscalls since last modification */
        unsigned long train_count;      /* # syscalls seen during training */
        u8 empty[PH_NUM_SYSCALLS];
        pH_seqflags entry[PH_NUM_SYSCALLS][PH_NUM_SYSCALLS]; 
} pH_disk_profile_data;

typedef struct pH_disk_profile {
        char magic[PH_FILE_MAGIC_LEN];  /* file magic: identifier, version */
        int normal;     /* Is this profile considered normal yet? */
        int frozen;
        time_t normal_time;
        int length;
        unsigned long count;      /* number of calls seen by this profile */
        int anomalies;  /* NOT ALF - used to decide if normal yet */
        pH_disk_profile_data train, test;
        char filename[PH_MAX_DISK_FILENAME];
} pH_disk_profile;

void usage(char *program)
{
        fprintf(stderr,
                "Usage: %s start-normal <profile-1> ... |\n"
                "       %s stop-normal <profile-1> ...\n",
                program, program);
}

int main(int argc, char *argv[])
{
        pH_disk_profile profile;
        int retval, fd, normal_val, file;

        if (argc < 3) {
                usage(argv[0]);
                return -1;
        }

        if (strcmp(argv[1], "start-normal") == 0)
                normal_val = 1;
        else if (strcmp(argv[1], "stop-normal") == 0)
                normal_val = 0;
        else {
                usage(argv[0]);
                return -1;
        }

        for (file=2; file < argc; file++) {
                fd = open(argv[file], O_RDONLY, 0);

                if (fd < 0) {
                        fprintf(stderr, "Can't open %s for reading\n",
                                argv[file]);
                        break;
                }
        
                retval = read(fd, &profile, sizeof(profile));

                close(fd);

                if (retval != sizeof(profile)) {
                        fprintf(stderr, "Only got %d bytes from %s, "
                                "but wanted %d\n", retval, argv[file],
                                sizeof(profile));
                        break;
                }

                if (strcmp(profile.magic, PH_FILE_MAGIC) !=0) {
                        fprintf(stderr, "Wrong magic for %s\n",
                                argv[file]);
                        break;
                }

                profile.normal = normal_val;

                if (profile.normal)
                        profile.anomalies = 0;

                fd = open(argv[file], O_WRONLY|O_TRUNC, 0);

                if (fd < 0) {
                        fprintf(stderr, "Can't open %s for writing\n",
                                argv[file]);
                        break;
                }

                retval = write(fd, &profile, sizeof(profile));
                close(fd);
                if (retval != sizeof(profile)) {
                        fprintf(stderr, "Only wrote %d bytes from %s, "
                                "but wanted %d\n", retval, argv[file],
                                sizeof(profile));
                        break;
                }

                printf("Updated %s\n", argv[file]);
        }

        return 0;
}
