/*
 *  pH-print-profile.c
 *
 *  Copyright (C) 2000, 2001  Anil Somayaji
 *
 *  A utility for printing the contents of a pH profile.  pH is a
 *  kernel extension for online process intrusion detection monitoring
 *  and response.
 *
 *  Licensed under the GNU GPL (see COPYING file), and as such has NO
 *  WARRANTY!  In particular, pH modifies the filesystem using
 *  low-level access routines.  Have a good backup, you could lose
 *  data through disk corruption!
 *
 *  August 16, 2000: pH-0.1, initial release
 *  July 17, 2001:   pH-0.17
 *
 *  $Id: pH-print-profile.c,v 1.20 2002/01/14 07:31:26 soma Exp $ */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <time.h>

#define PH_NUM_SYSCALLS 256
#define PH_MAX_SEQLEN 9
#define PH_MAX_FILENAME 255
#define PH_LOCALITY_WIN 128
#define PH_FILE_MAGIC_LEN 20
#define PH_MAX_DISK_FILENAME 256

const char *PH_FILE_MAGIC="pH profile 0.18\n";

typedef unsigned char u8;

typedef u8 pH_seqflags;

typedef struct pH_disk_profile_data {
        int sequences;
        unsigned long last_mod_count;/* # syscalls since last modification */
        unsigned long train_count;      /* # syscalls seen during training */
        u8 empty[PH_NUM_SYSCALLS];
        pH_seqflags entry[PH_NUM_SYSCALLS][PH_NUM_SYSCALLS]; 
} pH_disk_profile_data;

typedef struct pH_disk_profile {
        char magic[PH_FILE_MAGIC_LEN];  /* file magic: identifier, version */
        int normal;     /* Is this profile considered normal yet? */
        int frozen;
        time_t normal_time;
        int length;
        unsigned long count;      /* number of calls seen by this profile */
        int anomalies;  /* NOT ALF - used to decide if normal yet */
        pH_disk_profile_data train, test;
        char filename[PH_MAX_DISK_FILENAME];
} pH_disk_profile;

void print_pair(int current, int prev, int look)
{
        int i;

        printf("%4d", prev);
        for (i = 0; i < look; i++)
                printf(" *");
        printf(" %4d\n", current);
}


void print_pairs(pH_disk_profile *profile)
{
        int i,j,k;
        u8 train, test;
        
        printf("Training pairs:\n");

        for (k=0; k<8; k++) {
                for (i=0; i<PH_NUM_SYSCALLS; i++) {
                        for (j=0; j<PH_NUM_SYSCALLS; j++) {
                                train = profile->train.entry[i][j];
                                if ((train >> k) & ~(~0 << 1)) {
                                        print_pair(i, j, k);
                                }
                        }
                }
        }

        printf("\nTesting pairs:\n");

        for (k=0; k<8; k++) {
                for (i=0; i<PH_NUM_SYSCALLS; i++) {
                        for (j=0; j<PH_NUM_SYSCALLS; j++) {
                                test = profile->test.entry[i][j];
                                if ((test >> k) & ~(~0 << 1))
                                        print_pair(i, j, k);
                        }
                }
        }
}

int main(int argc, char *argv[])
{
        pH_disk_profile profile;
        int retval;
        int train_pairs;
        int test_pairs;
        int fd;
        int i,j,k,file;
        int num_profiles, num_normal, num_frozen;
        u8 train, test;
        int verbose = 0;
        int brief = 0;
        int file_start = 1;
        double total_calls;

        num_profiles = 0;
        num_normal = 0;
        num_frozen = 0;
        total_calls = 0;

        if (argc < 2)
                return 0;

        if (strcmp(argv[1], "-v") == 0) {
                verbose = 1;
                file_start = 2;
        } else if (strcmp(argv[1], "-b") == 0) {
                brief = 1;
                file_start = 2;
        }

        for (file=file_start; file < argc; file++) {
                fd = open(argv[file], O_RDONLY, 0);

                if (fd < 0) {
                        fprintf(stderr, "Can't open %s\n", argv[file]);
                        continue;
                }
        
                retval = read(fd, &profile, sizeof(profile));

                close(fd);

                if (retval != sizeof(profile)) {
                        fprintf(stderr, "Only got %d bytes from %s, "
                                "but wanted %d\n", retval, argv[file],
                                sizeof(profile));
                        continue;
                }

                if (strcmp(profile.magic, PH_FILE_MAGIC) !=0) {
                        fprintf(stderr, "Wrong magic for %s\n",
                                argv[file]);
                        continue;
                }

                train_pairs = 0;
                test_pairs = 0;
                for (i=0; i<PH_NUM_SYSCALLS; i++) {
                        for (j=0; j<PH_NUM_SYSCALLS; j++) {
                                train = profile.train.entry[i][j];
                                test = profile.test.entry[i][j];
                                for (k=0; k<8; k++) {
                                        if ((train >> k) & ~(~0 << 1))
                                                train_pairs++;
                                        if ((test >> k) & ~(~0 << 1))
                                                test_pairs++;
                                }
                        }
                }

                num_profiles++;
                if (profile.normal)
                        num_normal++;
                if (profile.frozen)
                        num_frozen++;

                total_calls +=profile.count;

                if (brief) {
                        printf("%10lu  %5d N%d  F%d  %5d %5d  %5d %5d %s\n",
                               profile.count,
                               profile.anomalies,
                               profile.normal,
                               profile.frozen,
                               train_pairs,
                               profile.train.sequences,
                               test_pairs,
                               profile.test.sequences,
                               profile.filename);
                } else {
                        printf("filename: %s\n"
                               "length:               %d\n"
                               "count:                %lu\n"
                               "normal:               %d\n"
                               "frozen:               %d\n"
                               "normal time:          %s"
                               "anomalies:            %d\n"
                               "train sequences:      %d\n"
                               "      pairs:          %d\n"
                               "      last_mod_count: %lu\n"
                               "      train_count:    %lu\n"
                               "test  sequences:      %d\n"
                               "      pairs:          %d\n"
                               "      last_mod_count: %lu\n"
                               "      train_count:    %lu\n"
                               "\n",
                               profile.filename,
                               profile.length,
                               profile.count,
                               profile.normal,
                               profile.frozen,
                               profile.normal_time ?
                               ctime(&(profile.normal_time)) : "none\n",
                               profile.anomalies,
                               profile.train.sequences,
                               train_pairs,
                               profile.train.last_mod_count,
                               profile.train.train_count,
                               profile.test.sequences,
                               test_pairs,
                               profile.test.last_mod_count,
                               profile.test.train_count);
                }
                
                if (verbose)
                        print_pairs(&profile);

        }
        
        printf("%d profiles total, %d normal, %d frozen, "
               "%.0f syscalls total.\n",
               num_profiles, num_normal, num_frozen, total_calls);

        return 0;
}
