/*
 *  pH-print-syscalls.c
 *
 *  Copyright (C) 2000  Anil Somayaji
 *
 *  A utility for printing the contents of a pH system call log.
 * 
 *  pH is a kernel extension for online process intrusion detection
 *  monitoring and response.
 *
 *  Licensed under the GNU GPL (see COPYING file), and as such has NO
 *  WARRANTY!  In particular, pH modifies the filesystem using
 *  low-level access routines.  Have a good backup, you could lose
 *  data through disk corruption!
 *
 *  July 17, 2001:      pH-0.17, initial release
 *
 *  $Id: pH-print-syscalls.c,v 1.12 2001/08/31 03:52:52 soma Exp $
 *
 */

#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <asm/unistd.h>
#include <stdio.h>
#include <stdlib.h>

#define PH_CALLREC_SYSCALL 0
#define PH_CALLREC_FORK    1
#define PH_CALLREC_EXECVE  2

#define ACTION_RAW 0
#define ACTION_ONE_INT 1
#define ACTION_ALL_INT 2
#define ACTION_STATS 3

#define MAX_SYSCALL 255

typedef unsigned char u8;
typedef unsigned short u16;

const int MAXPID = 40000;

typedef struct pH_call_logrec {
        u16 pid;
        union {
                u16 syscall;      /* type = 0 */
                u16 child_pid;    /* type = 1 (fork) */
                u16 filename_len;      /* type = 2 (execve) */
        } u;
        unsigned long count;
        long sec;
        long usec;
        u8 type;          /* 0 = regular call, 1 = fork, 2, execve */
} pH_call_logrec;

typedef struct program_rec program_rec;

struct program_rec {
        char *fn;
        FILE *int_file;
        int count;
        program_rec *next, *prev;
};


/* open_file: opens a file, and if necessary creates the
   leading path.  Always succeeds - otherwise, it just causes the
   program to exit.
*/

FILE *open_file(char *path)
{
        FILE *file;
        char buf[BUFSIZ];

        file = fopen(path, "a+");
        if (file)
                return file;

        snprintf(buf, BUFSIZ, "mkdir -p %s", path);
        system(buf);
        snprintf(buf, BUFSIZ, "rmdir %s", path);
        system(buf);

        file = fopen(path, "a+");
        if (file)
                return file;
        else {
                fprintf(stderr, "Cannot open %s\n", path);
                exit(-1);
                return NULL;
        }
}

void free_program(program_rec *program)
{
        program->count--;
        if (program->count <= 0) {
                if (program->prev) {
                        program->prev->next = program->next;
                }
                if (program->next) {
                        program->next->prev = program->prev;
                }
                fclose(program->int_file);
                free(program->fn);
                free(program);
        }
}

program_rec *get_program(program_rec *all_programs, char *fn)
{
        program_rec *program;

        program = all_programs;

        if (!program) {
                fprintf(stderr, "Null program list for %s!\n", fn);
                exit(-1);
        }

        while (program) {
                if (strcmp(program->fn, fn) == 0) {
                        program->count++;
                        return program;
                }
                program = program->next;
        }

        program = (program_rec *) malloc(sizeof(program_rec));
        program->fn = fn;
        program->int_file = open_file(fn);
        program->count = 1;

        program->next = all_programs->next;
        program->prev = all_programs;

        if (all_programs->next)
                all_programs->next->prev = program;
        all_programs->next = program;

        return program;
}

int write_ints(int fd, char *intdir)
{
        pH_call_logrec rec;
        int retval, prefix_len, fn_len, i;
        FILE *file;
        char *fn, buf[BUFSIZ];
        program_rec *pid[MAXPID];
        program_rec *all_programs, *unknown_program, *program;

        prefix_len = strlen(intdir);
        fn_len = 9 + prefix_len;
        fn = (char *) malloc(fn_len * (sizeof(char)));
        snprintf(fn, fn_len, "%s/UNKNOWN", intdir);
        file = open_file(fn);

        unknown_program = (program_rec *) malloc(sizeof(program_rec));
        unknown_program->fn = fn;
        fn = NULL;
        unknown_program->int_file = file;
        unknown_program->count = INT_MAX / 2;
        unknown_program->next = NULL;
        unknown_program->prev = NULL;

        for (i = 0; i < MAXPID; i++)
                pid[i] = unknown_program;

        all_programs = unknown_program;
        
        snprintf(buf, BUFSIZ, "%s/PIDLIST", intdir);
        file = fopen(buf, "r");
        if (file) {
                int p;
                
                while (!feof(file)) {
                        if (fscanf(file, "%d %s", &p, buf) == 2) {
                                fn_len = strlen(buf) + 1;
                                fn = (char *) malloc(fn_len * (sizeof(char)));
                                strcpy(fn, buf);
                                pid[p] = get_program(all_programs, fn);
                        }
                }
                fclose(file);
        }

        do {
                retval = read(fd, &rec, sizeof(rec));
                
                if (retval != sizeof(rec))
                        break;
                
                if (rec.pid >= MAXPID) {
                        fprintf(stderr, "pid %d too large!\n", rec.pid);
                        break;
                }

                if (rec.type == PH_CALLREC_EXECVE) {
                        fn_len = rec.u.filename_len + prefix_len + 5;
                        fn = (char *) malloc(fn_len * sizeof(char));
                        strcpy(fn, intdir);
                        retval = read(fd, fn + prefix_len,
                                      rec.u.filename_len);
                        if (retval == rec.u.filename_len) {
                                strcpy(fn + prefix_len + 
                                       strlen(fn + prefix_len),
                                       ".int");
                                program = pid[rec.pid];
                                free_program(program);
                                pid[rec.pid] = get_program(all_programs, fn);
                        } else {
                                fprintf(stderr, 
                                        "Bad exec filename at count %lu\n",
                                        rec.count);
                                free(fn);
                        }
                } else if (rec.type == PH_CALLREC_FORK) {
                        program = pid[rec.pid];
                        pid[rec.u.child_pid] = program;
                        program->count++;
                        fprintf(program->int_file,
                                "%hu fork\n", rec.u.child_pid);
                } else if (rec.type == PH_CALLREC_SYSCALL) {
                        program = pid[rec.pid];
                        fprintf(program->int_file,
                                "%hu %hu\n", rec.pid,
                                rec.u.syscall);
                        if (rec.u.syscall == __NR_exit) {
                                free_program(program);
                                pid[rec.pid] = unknown_program;
                                unknown_program->count++;
                        }
                } else {
                        fprintf(stderr, "Unknown record type %hhu\n",
                                rec.type);
                }
        } while (retval > 0);
        
        program = all_programs;
        while (program) {
                fclose(program->int_file);
                program = program->next;
        }

        fn_len = 9 + prefix_len;
        fn = (char *) malloc(fn_len * (sizeof(char)));
        snprintf(fn, fn_len, "%s/PIDLIST", intdir);
        file = fopen(fn, "w");        
        if (!file)
                return -1;
        for (i = 0; i < MAXPID; i++) {
                if (pid[i] != unknown_program) {
                        fprintf(file, "%d %s\n",
                                i, pid[i]->fn);
                }
        }
        fclose(file);
        free(fn);

        return 0;
}


void print_raw(int fd)
{
        pH_call_logrec rec;
        int retval;
        char fn[4096];

        do {
                retval = read(fd, &rec, sizeof(rec));
                
                if (retval == sizeof(rec)) {
                        if (rec.type == PH_CALLREC_EXECVE) {
                                retval = read(fd, fn, rec.u.filename_len);
                                if (retval == rec.u.filename_len)
                                        printf("EXEC pid %hu, count %lu: "
                                               "file %s\n", rec.pid, 
                                               rec.count, fn);
                                else
                                        fprintf(stderr, "Bad EXECVE rec\n");
                        } else if (rec.type == PH_CALLREC_FORK) {
                                printf("FORK pid %hu, count %lu: child %hu\n",
                                       rec.pid, rec.count, rec.u.child_pid);
                        } else if (rec.type == PH_CALLREC_SYSCALL) {
                                printf("%5hu  %3hu  %12lu  %ld.%ld\n",
                                       rec.pid,
                                       rec.u.syscall,
                                       rec.count,
                                       rec.sec,
                                       rec.usec);
                        } else {
                                fprintf(stderr, "Unknown record type %hhu\n",
                                        rec.type);
                        }
                }
                
        } while (retval > 0);
}

void print_stats(int fd)
{
        pH_call_logrec rec;
        int retval, i;
        char fn[4096];
        double call_count[MAX_SYSCALL + 1];
        
        for (i = 0; i <= MAX_SYSCALL; i++) {
                call_count[i] = 0;
        }

        do {
                retval = read(fd, &rec, sizeof(rec));
                
                if (retval == sizeof(rec)) {
                        if (rec.type == PH_CALLREC_EXECVE) {
                                retval = read(fd, fn, rec.u.filename_len);
                                if (retval != rec.u.filename_len)
                                        fprintf(stderr, "Bad EXECVE rec\n");
                        } else if (rec.type == PH_CALLREC_SYSCALL) {
                                if (rec.u.syscall <= MAX_SYSCALL) {
                                        call_count[rec.u.syscall]++;
                                } else {
                                        fprintf(stderr, "Call %hd too large "
                                                "at count %lu.\n",
                                                rec.u.syscall, rec.count);
                                }
                        }
                }
        } while (retval > 0);

        for (i = 0; i <= MAX_SYSCALL; i++) {
                printf("%4d %f\n", i, call_count[i]);
        }
}

void print_int(int fd, char *exec_fn)
{
        pH_call_logrec rec;
        int retval, i;
        char pid[MAXPID], fn[4096];

        for (i = 0; i < MAXPID; i++)
                pid[i] = 0;

        do {
                retval = read(fd, &rec, sizeof(rec));
                
                if (retval == sizeof(rec)) {
                        if (rec.type == PH_CALLREC_EXECVE) {
                                retval = read(fd, fn, rec.u.filename_len);
                                if (retval == rec.u.filename_len) {
                                        if (strcmp(exec_fn, fn) == 0)
                                                pid[rec.pid] = 1;
                                        else
                                                pid[rec.pid] = 0;
                                } else {
                                        fprintf(stderr, "Bad exec filename\n");
                                }
                        } else if (rec.type == PH_CALLREC_FORK) {
                                if (pid[rec.pid]) {
                                        pid[rec.u.child_pid] = 1;
                                        printf("%hu fork\n", rec.u.child_pid);
                                }
                        } else if (rec.type == PH_CALLREC_SYSCALL) {
                                if (pid[rec.pid]) {
                                        printf("%hu %hu\n", rec.pid,
                                               rec.u.syscall);
                                        if (rec.u.syscall == __NR_exit)
                                                pid[rec.pid] = 0;
                                }
                        } else {
                                fprintf(stderr, "Unknown record type %hhu\n",
                                       rec.type);
                        }
                }
                
        } while (retval > 0);
}


void usage(char *fn)
{
        fprintf(stderr, "Usage: %s int <logfile> <exec-filename> | \n", fn);
        fprintf(stderr, "       %s raw <logfile> | \n", fn);
        fprintf(stderr, "       %s stats <logfile> | \n", fn);
        fprintf(stderr, "       %s intdir <logfile> <int-dir> \n", fn);
        fprintf(stderr, "\nIf logfile is '-', "
                "then input is read from stdin\n");
}

int main(int argc, char *argv[])
{
        char *log_fn = NULL;
        int action, fd;

        if (argc < 3) {
                usage(argv[0]);
                return -1;
        }

        if ((strcmp(argv[1], "int") == 0) && (argc == 4)) {
                action = ACTION_ONE_INT;
        } else if (strcmp(argv[1], "raw") == 0) {
                action = ACTION_RAW;
        } else if (strcmp(argv[1], "stats") == 0) {
                action = ACTION_STATS;
        } else if ((strcmp(argv[1], "intdir") == 0) && (argc == 4)) {
                action = ACTION_ALL_INT;
        } else {
                usage(argv[0]);
                return -2;
        }
                       
        log_fn = argv[2];
        if (strcmp(log_fn, "-") == 0) {
                fd = 0;
        } else {
                fd = open(log_fn, O_RDONLY, 0);
                if (fd < 0) {
                        fprintf(stderr, "Can't open %s\n", log_fn);
                        return -1;
                }
        }

        switch (action) {
        case ACTION_ONE_INT :
                print_int(fd, argv[3]);
                break;
        case ACTION_RAW:
                print_raw(fd);
                break;
        case ACTION_STATS:
                print_stats(fd);
                break;
        case ACTION_ALL_INT:
                write_ints(fd, argv[3]);
                break;
        }
                
        close(fd);

        return 0;
}
